﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Mediachase.MetaDataPlus;
using Mediachase.MetaDataPlus.Configurator;
using System.Xml.Linq;
using Mediachase.Search.Extensions;

namespace Mediachase.Search.Providers.Solr35.Config
{
    internal static class SolrSchemaGenerator
    {
        #region Static Type Information

        private class SolrType
        {
            public string Name { get; private set; }
            public string TokenizedName { get; private set; }
            public bool MultiValued { get; private set; }

            public SolrType(string name, string tokenizedName, bool multiValued)
            {
                this.Name = name;
                this.TokenizedName = tokenizedName;
                this.MultiValued = multiValued;
            }
        }
        private static readonly Dictionary<MetaDataType, SolrType> TypeMapping;        
        static SolrSchemaGenerator()
        {
            SolrType _bool = new SolrType("boolean", null, false);
            SolrType _tint = new SolrType("tint", null, false);
            SolrType _tlong = new SolrType("tlong", null, false);
            SolrType _tfloat = new SolrType("tfloat", null, false);
            SolrType _tdouble = new SolrType("tdouble", null, false);
            SolrType _tdate = new SolrType("tdate", null, false);
            SolrType _string = new SolrType("string", "text", false);
            SolrType _string_email = new SolrType("string_email", null, false);
            SolrType _string_url = new SolrType("string_url", null, false);
            SolrType _string_html = new SolrType("string", "text_html", false);
            SolrType _multi_string = new SolrType("string", "text", true);

            TypeMapping = new Dictionary<MetaDataType, SolrType>();
            TypeMapping.Add(MetaDataType.BigInt, _tlong);
            TypeMapping.Add(MetaDataType.Bit, _bool);
            TypeMapping.Add(MetaDataType.Char, _string);
            TypeMapping.Add(MetaDataType.DateTime, _tdate);
            TypeMapping.Add(MetaDataType.Decimal, _tdouble);
            TypeMapping.Add(MetaDataType.Float, _tdouble);
            TypeMapping.Add(MetaDataType.Int, _tint);
            TypeMapping.Add(MetaDataType.Money, _tdouble);
            TypeMapping.Add(MetaDataType.NChar, _string);
            TypeMapping.Add(MetaDataType.NText, _string);
            TypeMapping.Add(MetaDataType.NVarChar, _string);
            TypeMapping.Add(MetaDataType.Real, _tfloat);
            TypeMapping.Add(MetaDataType.UniqueIdentifier, _string);
            TypeMapping.Add(MetaDataType.SmallDateTime, _tdate);
            TypeMapping.Add(MetaDataType.SmallInt, _tint);
            TypeMapping.Add(MetaDataType.SmallMoney, _tfloat);
            TypeMapping.Add(MetaDataType.Text, _string);
            TypeMapping.Add(MetaDataType.TinyInt, _tint);
            TypeMapping.Add(MetaDataType.VarChar, _string);
            TypeMapping.Add(MetaDataType.Numeric, _tdouble);
            TypeMapping.Add(MetaDataType.Sysname, _string);
            TypeMapping.Add(MetaDataType.Integer, _tint);
            TypeMapping.Add(MetaDataType.Boolean, _bool);
            TypeMapping.Add(MetaDataType.Date, _tdate);
            TypeMapping.Add(MetaDataType.Email, _string_email);
            TypeMapping.Add(MetaDataType.URL, _string_url);
            TypeMapping.Add(MetaDataType.ShortString, _string);
            TypeMapping.Add(MetaDataType.LongString, _string);
            TypeMapping.Add(MetaDataType.LongHtmlString, _string_html);
            TypeMapping.Add(MetaDataType.DictionarySingleValue, _string);
            TypeMapping.Add(MetaDataType.DictionaryMultiValue, _multi_string);
            TypeMapping.Add(MetaDataType.EnumSingleValue, _string);
            TypeMapping.Add(MetaDataType.EnumMultiValue, _multi_string);
            TypeMapping.Add(MetaDataType.StringDictionary, _multi_string);            
        }

        #endregion


        public static string GenerateCatalogConfiguration(MetaDataContext catalogContext)
        {
            // metafield names arent necessarily unique, even if we didnt lowercase them.
            // make sure we handle conflicts cleanly.
            Dictionary<string, List<MetaField>> fieldGroups = GetCatalogEntryMetaFields(catalogContext)
                .Where(f => IncludeFieldInConfig(f))
                .GroupBy(f => f.Name.ToLowerInvariant())
                .ToDictionary(grp => grp.Key, grp => grp.ToList());

            StringBuilder builder = new StringBuilder();
            builder.AppendLine("<!-- Begin Generated schema.xml Content -->");
            AppendFieldConfiguration(builder, fieldGroups);
            builder.AppendLine("  </fields>");
            AppendCopyFieldConfiguration(builder, fieldGroups);
            builder.AppendLine("<!-- End Generated schema.xml Content -->");

            return builder.ToString();
        }


        private static void AppendFieldConfiguration(StringBuilder builder, Dictionary<string, List<MetaField>> fieldGroups)
        {
            foreach (KeyValuePair<string, List<MetaField>> fieldGroup in fieldGroups
                .OrderBy(kv => kv.Key)
                .Where(kv => !string.Equals(kv.Key, "displayname", StringComparison.OrdinalIgnoreCase)))
            {
                string fieldName = fieldGroup.Key;
                List<bool> indexValues = fieldGroup.Value.Select(f => SolrIndexField(f)).Distinct().ToList();
                List<bool> storeValues = fieldGroup.Value.Select(f => SolrStoreField(f)).Distinct().ToList();
                List<string> fieldTypeValues = fieldGroup.Value
                    .Select(f => SolrTokenizeField(f) ? TypeMapping[f.DataType].TokenizedName : TypeMapping[f.DataType].Name)
                    .Distinct()
                    .ToList();
                List<bool> multiValuedValues = fieldGroup.Value
                    .Select(f => TypeMapping[f.DataType].MultiValued)
                    .Distinct()
                    .ToList();
                List<bool> includeInContentValues = fieldGroup.Value
                    .Select(f => SolrIncludeInDefaultSearch(f))
                    .Distinct()
                    .ToList();

                if (fieldTypeValues.Count > 1 || multiValuedValues.Count > 1)
                {
                    AppendFieldConflictError(builder, fieldName, fieldTypeValues, indexValues, storeValues, multiValuedValues);
                }
                else
                {
                    // if there are multiple metafields with the same name, and any one of them has index="true", use index="true". same for store.
                    bool index = indexValues.Any(b => b);
                    bool store = storeValues.Any(b => b);

                    if (indexValues.Count > 1 || storeValues.Count > 1) AppendFieldConflictWarning(builder, fieldName, indexValues.Count > 1, storeValues.Count > 1);
                        
                    builder.Append("    ");
                    AppendFieldElement(builder, fieldName, fieldTypeValues.Single(), indexValues.Any(b => b), storeValues.Any(b => b), multiValuedValues.Single());                    
                }
            }

            builder.AppendLine("    <dynamicField name=\"*\" type=\"ignored\"/>");
        }


        private static void AppendCopyFieldConfiguration(StringBuilder builder, Dictionary<string, List<MetaField>> fieldGroups)
        {
            AppendCopyFieldElement(builder.Append("  "), "code");

            foreach (KeyValuePair<string, List<MetaField>> fieldGroup in fieldGroups.OrderBy(kv => kv.Key))
            {
                string fieldName = fieldGroup.Key;
                List<bool> includeInContentValues = fieldGroup.Value.Select(f => SolrIncludeInDefaultSearch(f)).Distinct().ToList();

                if (includeInContentValues.Count > 1) AppendCopyFieldConflictWarning(builder, fieldName);
                
                if (includeInContentValues.Any(b => b)) AppendCopyFieldElement(builder.Append("  "), fieldName);
            }
        }


        private static void AppendFieldElement(StringBuilder builder, string fieldName, string fieldType, bool indexed, bool stored, bool multiValued)
        {
            XElement field = new XElement("field",
                new XAttribute("name", fieldName),
                new XAttribute("type", fieldType),
                new XAttribute("indexed", indexed ? "true" : "false"),
                new XAttribute("stored", stored ? "true" : "false"),
                new XAttribute("multiValued", multiValued ? "true" : "false"));
            builder.AppendLine(field.ToString());
        }


        private static void AppendCopyFieldElement(StringBuilder builder, string fieldName)
        {
            XElement copyField = new XElement("copyField",
                new XAttribute("source", fieldName),
                new XAttribute("dest", "_content"));
            builder.AppendLine(copyField.ToString());
        }


        private static void AppendFieldConflictError(StringBuilder builder, string fieldName, List<string> fieldTypeValues, List<bool> indexValues, List<bool> storeValues, List<bool> multiValuedValues)
        {
            builder
                .AppendFormat("    <!-- ERROR: There are multiple catalog entry metafields named '{0}' with conflicting indexing configuration.", fieldName)
                .AppendLine()
                .AppendFormat("         Metafields named '{0}' will not be indexed or stored unless a matching field element is manually added.", fieldName)
                .AppendLine();

            builder.AppendLine("        Possible configurations:");
                
            foreach (string fieldType in fieldTypeValues.OrderBy(x => x))
            {
                foreach (bool index in indexValues.OrderBy(x => x))
                {
                    foreach (bool store in indexValues.OrderBy(x => x))
                    {
                        foreach (bool multiValued in multiValuedValues.OrderBy(x => x))
                        {
                            AppendFieldElement(builder.Append("        "), fieldName, fieldType, index, store, multiValued);
                            builder.AppendLine();
                        }
                    }
                }
            }

            builder.AppendLine("    -->");
        }
        
    
        private static void AppendFieldConflictWarning(StringBuilder builder, string fieldName, bool indexConflict, bool storeConflict)
        {
            builder
                .AppendFormat( "    <!-- WARNING: There are multiple catalog entry metafields named '{0}' with conflicting indexing configurations.", fieldName)
                .AppendLine();
                            
            if (indexConflict && storeConflict)
            {
                builder
                    .AppendFormat("        The values of all metafields named '{0}' will be indexed and stored, even though some metafields with this name are configured not to be.", fieldName)
                    .AppendLine();
            }
            else if (indexConflict)
            {
                builder
                    .AppendFormat("        The values of all metafields named '{0}' will be indexed, even though some metafields with this name are configured not to be.", fieldName)
                    .AppendLine();
            }
            else if (storeConflict)
            {
                builder
                    .AppendFormat("        The values of all metafields named '{0}' will be stored, even though some metafields with this name are configured not to be.", fieldName)
                    .AppendLine();
            }

            builder.AppendLine("    -->");
        }


        private static void AppendCopyFieldConflictWarning(StringBuilder builder, string fieldName)
        {
            builder
                .AppendFormat("    <!-- WARNING: There are multiple catalog entry metafields named '{0}' with conflicting indexing configurations.", fieldName)
                .AppendLine()
                .AppendFormat("        The values of all metafields named '{0}' will be added to the _content field, even though some metafields with this name are configured not to be.", fieldName)
                .AppendLine(  "    -->");
        }


        #region Typing Judgements

        private static bool IncludeFieldInConfig(MetaField field)
        {
            return SolrIndexField(field) || SolrStoreField(field) || SolrIncludeInDefaultSearch(field);
        }

        private static bool SolrIndexField(MetaField field)
        {
            return (field.SafeAllowSearch || field.IsIndexSortable()) &&
                TypeMapping.ContainsKey(field.DataType);
        }

        private static bool SolrStoreField(MetaField field)
        {
            return field.IsIndexStored() && TypeMapping.ContainsKey(field.DataType);
        }

        private static bool SolrTokenizeField(MetaField field)
        {
            SolrType solrType;
            return
                field.IsIndexTokenized() &&
                TypeMapping.TryGetValue(field.DataType, out solrType) &&
                solrType.TokenizedName != null;
        }

        private static bool SolrIncludeInDefaultSearch(MetaField field)
        {
            return
                field.IsIncludedInDefaultSearch() &&
                TypeMapping.ContainsKey(field.DataType);
        }

        #endregion


        /// <summary>
        /// Enumerates the distinct metafields attached to the catalog entry metaclass or any descendant of the catalog entry metaclass.
        /// </summary>
        /// <param name="catalogContext">The catalog system metadata context.</param>
        /// <returns>An enumeration of distinct metafields.</returns>
        private static IEnumerable<MetaField> GetCatalogEntryMetaFields(MetaDataContext catalogContext)
        {
            Queue<string> classNameQueue = new Queue<string>();
            HashSet<int> processedClasses = new HashSet<int>();
            HashSet<int> processedFields = new HashSet<int>();

            classNameQueue.Enqueue("CatalogEntry");
            
            while (classNameQueue.Count > 0)
            {
                var className = classNameQueue.Dequeue();
                MetaClass mclass = MetaClass.Load(catalogContext, className);
                if (mclass != null && !processedClasses.Contains(mclass.Id))
                {
                    processedClasses.Add(mclass.Id);
                    foreach (MetaClass childClass in mclass.ChildClasses)
                    {
                        classNameQueue.Enqueue(childClass.Name);
                    }

                    foreach (MetaField field in mclass.MetaFields)
                    {
                        if (!processedFields.Contains(field.Id))
                        {
                            processedFields.Add(field.Id);
                            yield return field;
                        }
                    }
                }
            }
        }

    }
}
