﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Xml.Linq;
using Mediachase.Search.Extensions;

namespace Mediachase.Search.Providers.Solr35.Select
{
    internal static class QueryResponseParser
    {
        /// <summary>
        /// Parses a search response from solr.
        /// </summary>
        /// <exception cref="System.FormatException">
        /// Thrown if response text is not an XML document with at least one element
        /// matching /reponse/result[@name="response" and @numFound], or if the
        /// value of the numFound attribute of that element is not an integer.
        /// </exception>
        /// <param name="criteria">The search criteria that was used to create the request.</param>
        /// <param name="query">The query object representing the request.</param>
        /// <param name="response">The body of the response from the Solr instance.</param>
        /// <param name="explicitFieldFacetValuesOnly">If true, only values explicitly included for field facets will be included in the response; otherwise, all values with counts greater than 0 will be included.</param>
        /// <returns>An ISearchResults object representing the results of the search.</returns>
        public static ISearchResults Parse(ISearchCriteria criteria, ISolrSearchQuery query, string response, bool explicitFieldFacetValuesOnly)
        {
            SearchResults results = new SearchResults(new SearchDocuments(), criteria);

            XDocument xml = XDocument.Parse(response);
            XElement responseElement = xml.Elements("response").Single();
            XElement resultElement = responseElement.FirstChildOrDefault("result", "response");
            XAttribute numFoundAttribute = resultElement.Attribute("numFound");

            if (numFoundAttribute != null) results.Documents.TotalCount = int.Parse(resultElement.Attribute("numFound").Value);
            else throw new FormatException("Required attribute /response/result[@name=\"response\"]/@numFound not present.");

            // get documents.
            foreach (XElement docElement in resultElement.Elements("doc"))
            {
                SearchDocument doc = new SearchDocument();
                foreach (ISearchField field in ParseFields(docElement)) doc.Add(field);
                results.Documents.Add(doc);
            }

            // get facet counts.
            XElement facetCountsElement = responseElement.FirstChildOrDefault("lst", "facet_counts");
            if (facetCountsElement != null)
            {
                var facetmanger = query.Faceting as FacetManager;
                if (facetmanger == null)
                {
                    results.FacetGroups = query.Faceting.BuildResponse(
                        ParseFieldCounts(facetCountsElement),
                        ParseQueryCounts(facetCountsElement),
                        criteria.Locale,
                        explicitFieldFacetValuesOnly).ToArray();
                }
                else
                {
                    results.FacetGroups = facetmanger.BuildResponse(
                        ParseFieldCounts(facetCountsElement),
                        ParseQueryCounts(facetCountsElement),
                        criteria,
                        explicitFieldFacetValuesOnly).ToArray();
                    
                }
            }

            // get spellcheck suggestions.
            // note: this is unlikely to ever result in data; the query builder does not support specifying spellcheck
            // query options.  it may be possible to force spellcheck on all queries in the solr configuration.
            XElement spellCheckElement = responseElement.FirstChildOrDefault("lst", "spellcheck");
            if (spellCheckElement != null)
            {
                XElement suggestionsElement = spellCheckElement.FirstChildOrDefault("lst", "suggestions");
                if (spellCheckElement != null)
                {
                    results.Suggestions = suggestionsElement
                        .Elements("lst")
                        .Select(lst => lst.FirstChildOrDefault("arr", "suggestion"))
                        .SelectMany(arr => arr.Elements("str").Select(str => str.Value))
                        .ToArray();
                }
            }

            return results;
        }


        /// <summary>
        /// Gets the first child of <paramref name="element"/> that is an element of type
        /// <paramref name="elementName"/>, and has a name attribute equal to
        /// <paramref name="attributeNameValue"/>.
        /// </summary>
        /// <param name="element">The element to search within.</param>
        /// <param name="elementName">The element type to search for.</param>
        /// <param name="attributeNameValue">The value to search the name attribute for.</param>
        /// <returns>The first element matching the parameters, or null if none is found.</returns>
        private static XElement FirstChildOrDefault(this XElement element, string elementName, string attributeNameValue)
        {
            foreach (XElement el in element.Elements(elementName))
            {
                XAttribute name = el.Attribute("name");
                if (name != null && name.Value == attributeNameValue) return el;
            }

            return null;
        }


        /// <summary>
        /// Gets all children of <paramref name="element"/> that are elements of type
        /// <paramref name="elementName"/>, and have name attributes.
        /// </summary>
        /// <param name="element">The element to search within.</param>
        /// <param name="elementName">The element type to search for.</param>
        /// <returns>An enumeration of the matching elements.</returns>
        private static IEnumerable<XElement> ChildrenWithNames(this XElement element, string elementName)
        {
            return element
                .Elements(elementName)
                .Where(el => el.Attribute("name") != null);
        }


        /// <summary>
        /// Parses the fields in a doc element.
        /// </summary>
        /// <remarks>
        /// Any fields that cannot be parsed are silently ignored.
        /// </remarks>
        /// <param name="docElement">A doc element in the search results.</param>
        /// <returns>An enumeration of successfully parsed fields.</returns>
        private static IEnumerable<ISearchField> ParseFields(XElement docElement)
        {
            foreach (XElement fieldElement in docElement.Elements())
            {
                ISearchField field = null;
                string name = fieldElement.Attribute("name").Value;

                if (fieldElement.Name == "arr")
                {
                    object[] values = fieldElement.Elements().Select(el => ParseValue(el)).Where(v => v != null).ToArray();
                    if (values.Length > 0) field = new SearchField(name, values);
                }
                else
                {
                    object value = ParseValue(fieldElement);
                    if (value != null) field = new SearchField(name, ParseValue(fieldElement));
                }

                if (field != null) yield return field;
            }
        }


        /// <summary>
        /// Parses a value element.
        /// </summary>
        /// <param name="valueElement">The element to parse.</param>
        /// <returns>An object containing the parsed value, or null if the value could not be parsed.</returns>
        private static object ParseValue(XElement valueElement)
        {
            XName type = valueElement.Name;
            string value = valueElement.Value;

            if (type == "str") return value;
            else if (type == "int")
            {
                int intValue;
                if (int.TryParse(value, out intValue)) return intValue;
                else return null;
            }
            else if (type == "bool")
            {
                bool boolValue;
                if (bool.TryParse(value, out boolValue)) return boolValue;
                else return null;
            }
            else if (type == "date")
            {
                DateTime dateValue;
                if (DateTime.TryParseExact(
                    value,
                    SolrSearchProvider.DateTimeFormatString,
                    CultureInfo.InvariantCulture,
                    DateTimeStyles.None,
                    out dateValue)) return dateValue;
                else return null;
            }
            else if (type == "float")
            {
                double doubleValue;
                if (double.TryParse(value, out doubleValue)) return doubleValue;
                else return null;
            }
            else return null;
        }


        /// <summary>
        /// Parses facet field counts into response records for <see cref="FacetManager"/>.
        /// </summary>
        /// <param name="element">The facet counts element to parse.</param>
        /// <returns>An enumeration of <see cref="FacetManager.FieldResponse"/> objects.</returns>
        private static IEnumerable<IFacetFieldResponse> ParseFieldCounts(XElement element)
        {
            XElement facetFieldsElement = element.FirstChildOrDefault("lst", "facet_fields");
            if (facetFieldsElement != null)
            {
                foreach (XElement fieldElement in facetFieldsElement.ChildrenWithNames("lst"))
                {
                    string field = fieldElement.Attribute("name").Value;

                    foreach (XElement countElement in fieldElement.ChildrenWithNames("int"))
                    {
                        int count;
                        if (int.TryParse(countElement.Value, out count))
                        {
                            yield return new FacetManager.FieldResponse(field, countElement.Attribute("name").Value, count);
                        }
                    }
                }
            }
        }


        /// <summary>
        /// Parses facet query counts into response records.
        /// </summary>
        /// <param name="element">The facet counts element to parse.</param>
        /// <returns>An enumeration of <see cref="FacetManager.QueryResponse"/> objects.</returns>
        private static IEnumerable<IFacetQueryResponse> ParseQueryCounts(XElement element)
        {
            XElement facetQueriesElement = element.FirstChildOrDefault("lst", "facet_queries");
            if (facetQueriesElement != null)
            {
                foreach (XElement countElement in facetQueriesElement.ChildrenWithNames("int"))
                {
                    int count;
                    if (int.TryParse(countElement.Value, out count))
                    {
                        yield return new FacetManager.QueryResponse(countElement.Attribute("name").Value, count);
                    }
                }
            }
        }
    }
}
