﻿using System;
using System.IO;
using System.Net;
using System.Text;
using System.Web;
using System.Web.Caching;

namespace Mediachase.Search.Providers.Solr35
{
    internal static class SolrHttp
    {
        internal class Response
        {
            public string ETag { get; private set; }
            public string Content { get; private set; }
            public object ProcessedObject { get; set; }

            public Response(string eTag, string content)
            {
                this.ETag = eTag;
                this.Content = content;
            }
        }

        internal class ResponseCache
        {
            private const string CacheKeyPrefixFormat = "solr{0}";
            private static readonly TimeSpan SlidingExpiration = new TimeSpan(0, 10, 0);

            private readonly string cacheKeyPrefix;

            public ResponseCache()
            {
                cacheKeyPrefix = string.Format("solr{0}", Guid.NewGuid());
            }

            public SolrHttp.Response this[string url]
            {
                get
                {
                    string cacheKey = GetCacheKey(url);
                    return HttpRuntime.Cache[cacheKey] as SolrHttp.Response;
                }
                set
                {
                    string cacheKey = GetCacheKey(url);
                    HttpRuntime.Cache.Insert(cacheKey, value, null, Cache.NoAbsoluteExpiration, SlidingExpiration);
                }
            }

            private string GetCacheKey(string url)
            {
                return cacheKeyPrefix + url;
            }
        }


        /// <summary>
        /// Executes the specified HTTP GET request.  If <paramref name="cache"/> is not null, cached entries
        /// will remit the cached request's ETag and appropriately handle a HTTP 304/Not Modified response.
        /// </summary>
        /// <param name="url">The URL to get.</param>
        /// <param name="cache">If non-null, a <see cref="ResponseCache"/> object used to enable and handle HTTP 304 responses.</param>
        /// <returns>An object representing the response.  If <paramref name="cache"/> is not null, the response may also be cached.</returns>
        public static Response Get(string url, ResponseCache cache)
        {
            HttpWebRequest request = (HttpWebRequest)WebRequest.Create(url);
            request.Method = "GET";
            request.KeepAlive = true;
            request.AutomaticDecompression = DecompressionMethods.Deflate | DecompressionMethods.GZip;

            Response cached = null;
            if (cache != null)
            {
                cached = cache[url];
                if (cached != null) request.Headers.Add(HttpRequestHeader.IfNoneMatch, cached.ETag);
            }

            Response response = null;
            try
            {
                using (HttpWebResponse httpResponse = (HttpWebResponse)request.GetResponse())
                {
                    string etag = null;
                    if (cache != null)
                    {
                        string cacheControl = httpResponse.Headers[HttpResponseHeader.CacheControl];
                        if (cacheControl == null || cacheControl.Contains("no-cache")) etag = null;
                        else etag = httpResponse.Headers[HttpResponseHeader.ETag];
                    }

                    response = new Response(etag, GetContent(httpResponse));

                    if (cache != null && etag != null) cache[url] = response;
                }
            }
            catch (WebException e)
            {
                HttpWebResponse httpResponse = e.Response as HttpWebResponse;
                if (httpResponse != null)
                {
                    if (httpResponse.StatusCode == HttpStatusCode.NotModified && cached != null) response = cached;
                    else throw;
                }                
            }

            return response;
        }


        /// <summary>
        /// Executes the specified HTTP POST request and returns the response as a string.
        /// </summary>
        /// <param name="url">The URL to post to.</param>
        /// <param name="content">The content to post.  The content will be encoded as UTF-8.</param>
        /// <param name="contentType">The type of the content.</param>
        /// <param name="cache">If non-null, a <see cref="ResponseCache"/> object used to enable and handle HTTP 304 responses.</param>        
        /// <returns>An object representing the response.  If <paramref name="cache"/> is not null, the response may also be cached.</returns>
        public static Response Post(string url, string content, string contentType, ResponseCache cache)
        {
            HttpWebRequest request = (HttpWebRequest)WebRequest.Create(url);
            request.Method = "POST";
            request.KeepAlive = true;
            request.AutomaticDecompression = DecompressionMethods.Deflate | DecompressionMethods.GZip;
            request.ContentType = contentType;
            request.ProtocolVersion = HttpVersion.Version11;

            byte[] contentBytes = Encoding.UTF8.GetBytes(content);
            request.ContentLength = contentBytes.Length;

            Response cached = null;
            if (cache != null)
            {
                cached = cache[url];
                if (cached != null) request.Headers.Add(HttpRequestHeader.IfNoneMatch, cached.ETag);
            }

            Response response = null;
            try
            {
                using (Stream postStream = request.GetRequestStream())
                {
                    postStream.Write(contentBytes, 0, contentBytes.Length);
                }

                using (HttpWebResponse httpResponse = (HttpWebResponse)request.GetResponse())
                {
                    string etag = null;
                    if (cache != null)
                    {
                        string cacheControl = httpResponse.Headers[HttpResponseHeader.CacheControl];
                        if (cacheControl == null || cacheControl.Contains("no-cache")) etag = null;
                        else etag = httpResponse.Headers[HttpResponseHeader.ETag];
                    }

                    response = new Response(etag, GetContent(httpResponse));

                    if (cache != null && etag != null) cache[url] = response;
                }
            }
            catch (WebException e)
            {
                HttpWebResponse httpResponse = e.Response as HttpWebResponse;
                if (httpResponse != null && httpResponse.StatusCode == HttpStatusCode.NotModified && cached != null)
                {
                    response = cached;
                }
                else
                {
                    // if the exception has response text, wrap it up as the message.
                    // if not, rethrow the exception as is.
                    Exception wrappedException = null;
                    try
                    {
                        string responseContent = GetContent(e.Response);
                        wrappedException = new Exception(responseContent, e);
                    }
                    catch { }

                    if (wrappedException != null) throw wrappedException;
                    else throw;
                }
            }

            return response;
        }


        private static string GetContent(WebResponse response)
        {
            string content;

            Encoding encoding = Encoding.UTF8;
            try 
            { 
                HttpWebResponse webResponse = response as HttpWebResponse;
                if (webResponse != null) encoding = Encoding.GetEncoding(webResponse.CharacterSet);
            }
            catch (ArgumentException) 
            {
                // supress ArgmentException from GetEncoding, indicating that the character set is not known/supported.
                // this results in the default Encoding.UTF8 value being used.
            }

            using (Stream stream = response.GetResponseStream())
            {
                using (StreamReader reader = new StreamReader(stream, encoding))
                {
                    content = reader.ReadToEnd();
                }
            }

            return content;
        }
    }
}
