﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Configuration.Provider;
using System.Linq;
using System.Text;
using System.Web;
using Mediachase.MetaDataPlus;
using Mediachase.Search.Extensions;
using Mediachase.Search.Providers.Solr35.Config;
using Mediachase.Search.Providers.Solr35.Select;
using Mediachase.Search.Providers.Solr35.Update;

namespace Mediachase.Search.Providers.Solr35
{
    public class SolrSearchProvider : SearchProvider, ISearchProviderConfigGenerator
    {
        #region Static Constants and Methods

        public const string DateTimeFormatString = "yyyy-MM-dd'T'HH:mm:ss'Z'";

        /// <summary>
        /// Gets the best matching description for a locale.
        /// </summary>
        /// <remarks>
        /// From highest priority to lowest, the attempted matches are:
        /// <list type="number">
        /// <item>A description with a locale that matches the specified locale exactly.</item>
        /// <item>A description with a locale that matches the specified locale's language only.</item>
        /// <item>A description with a locale that matches the description's default locale exactly.</item>
        /// <item>A description with a locale that matches the description's default locale's language only.</item>
        /// </list>
        /// All comparison are performed with StringComparison.OrdinalIgnoreCase.  If none of the above matches are found, null is returned.
        /// </remarks>
        /// <param name="descriptions">The object containing localized descriptions.</param>
        /// <param name="forLocale">The desired locale.</param>
        /// <returns>The description best matching the requested locale, or null if no matching description is found.</returns> 
        public static string GetLocalizedDescription(Descriptions descriptions, string forLocale)
        {
            Description match;

            match = descriptions.Description.FirstOrDefault(d => d.locale.Equals(forLocale, StringComparison.OrdinalIgnoreCase));
            if (match == null)
            {
                match = descriptions.Description.FirstOrDefault(d => MatchesLanguage(d.locale, forLocale));
                if (match == null)
                {
                    match = descriptions.Description.FirstOrDefault(d => d.locale.Equals(descriptions.defaultLocale, StringComparison.OrdinalIgnoreCase));
                    if (match == null)
                    {
                        match = descriptions.Description.FirstOrDefault(d => MatchesLanguage(d.locale, descriptions.defaultLocale));
                    }
                }
            }

            return match != null ? match.Value : null;
        }

        private static bool MatchesLanguage(string localeA, string localeB)
        {
            int minLength = localeA.Length < localeB.Length ? localeA.Length : localeB.Length;

            for (int i = 0; i < minLength; ++i)
            {
                if (i > 0 && localeA[i] == '-' && localeB[i] == '-') return true;
                else if (char.ToLowerInvariant(localeA[i]) != char.ToLowerInvariant(localeB[i])) return false;
            }

            if (localeA.Length > minLength) return localeA[minLength] == '-';
            else if (localeB.Length > minLength) return localeB[minLength] == '-';
            else return true;
        }

        #endregion


        #region Private Fields

        private static object updateBatchesLock = new object();
        private Dictionary<string, Dictionary<string, UpdateBatch>> updateBatches;

        public SolrSearchProvider()
        {
            updateBatches = new Dictionary<string, Dictionary<string, UpdateBatch>>();
        }

        #endregion


        #region Configuration Properties and Methods

        private ProviderConfiguration Config { get; set; }

        public override string QueryBuilderType
        {
            get { return Config == null ? null : Config.QueryBuilder.GetType().Name; }
        }

        private class ProviderConfiguration
        {
            /// <summary>
            /// An instance of the configuration query builder, from the "queryBuilderType" provider attribute. Required.
            /// </summary>
            public ISearchQueryBuilder QueryBuilder { get; private set; }

            /// <summary>
            /// The solr server URL. Specified in the "url" provider attribute. Required.
            /// </summary>
            public string SolrServerUrl { get; set; }

            /// <summary>
            /// If true, a single instance is used for all applications; if false, each application will use its own instance. 
            /// Specified in the "shareCores" provider attribute. Required.
            /// </summary>
            public bool ShareCores { get; set; }

            /// <summary>
            /// If true, queries are sent via http post; if false, by http get.
            /// Specified in the "queryMethod" provider attribute, which should be set to "get" or "post". Defaults to "post" (true) if not configured.
            /// </summary>
            public bool QueryUsingPost { get; set; }

            /// <summary>
            /// The commit-within window passed to the search engine, describing the maximum latency between
            /// an indexing message being sent to the server and the commit operation for that message beginning.
            /// The time is described in milliseconds.
            /// Specified in the "commitWithin" provider attribute. Defaults to 10000 if not configured.
            /// </summary>
            public int CommitWithin { get; set; }

            /// <summary>
            /// The maximum number of search documents, or document deletions, to send to the search engine in a single batch.
            /// Note that sending a batch does not necessarily commit the batch, commits are controlled with CommitWithin.
            /// Specified in the "maximumBatchSize" provider attribute. Defaults to 100 if not configured.
            /// </summary>
            public int MaximumBatchSize { get; set; }

            /// <summary>
            /// If true, field facet counts are only returned for values that are explicitly specified in the search criteria with
            /// a localizable description.  If false, field facet counts for all values are returned; note that unspecified values
            /// will be in lower case.
            /// Specified in the "facetLocalizedFieldValuesOnly" provider attribute. Defaults to true if not configured.
            /// </summary>
            public bool FacetLocalizedFieldValuesOnly { get; set; }

            /// <summary>
            /// If false, a filter will be applied to all searches to ensure the current date and time is between the catalog start and end date.
            /// If true, this filter will not be added.
            /// Specified in the "disableDateFilter" provider attribute.  Defaults to false if not configured.
            /// </summary>
            public bool DisableDateFilter { get; set; }

            /// <summary>
            /// A response cache to enable the sarch engine caching features.  If caching is not enabled, this will be null.
            /// Caching is enabled in the "enableCaching" provider attribute, which defaults to false if not configured.
            /// </summary>
            public SolrHttp.ResponseCache ResponseCache { get; set; }

            public ProviderConfiguration(string name, NameValueCollection config)
            {
                SolrServerUrl = TakeString(config, "url", true);
                if (!SolrServerUrl.EndsWith("/")) SolrServerUrl = SolrServerUrl + "/";

                SetQueryBuilder(config);

                bool enableCaching = TakeBool(config, "enableCaching", false) ?? false;
                ResponseCache = enableCaching ? new SolrHttp.ResponseCache() : null;

                QueryUsingPost = true; // default
                string queryPostString = TakeString(config, "queryMethod", false);
                if (queryPostString != null)
                {
                    queryPostString = queryPostString.Trim();
                    if (string.Equals(queryPostString, "get", StringComparison.OrdinalIgnoreCase)) QueryUsingPost = false;
                    else if (string.Equals(queryPostString, "post", StringComparison.OrdinalIgnoreCase)) QueryUsingPost = true;
                    else throw new Exception("queryMethod attribute could not be parsed.");
                }
                
                ShareCores = TakeBool(config, "shareCores", true).Value;
                CommitWithin = TakeInt(config, "commitWithin", false) ?? 10000;
                MaximumBatchSize = TakeInt(config, "maximumBatchSize", false) ?? 100;
                FacetLocalizedFieldValuesOnly = TakeBool(config, "facetLocalizedFieldValuesOnly", false) ?? true;
                DisableDateFilter = TakeBool(config, "disableDateFilter", false) ?? false;

                if (config.Count > 0)
                {
                    throw new ProviderException(string.Format("Unrecognized attribute \"{0}\" in configuration", config.GetKey(0)));
                }
            }

            private string TakeString(NameValueCollection config, string key, bool required)
            {
                string value = config[key];
                if (string.IsNullOrEmpty(value))
                {
                    if (required) throw new ProviderException(string.Format("{0} is a required attribute for SolrSearchProvider.", key));
                    else value = null;
                }

                config.Remove(key);
                
                return value;
            }

            private bool? TakeBool(NameValueCollection config, string key, bool required)
            {
                string stringValue = config[key];
                bool boolValue;
                bool? result = null;

                if (string.IsNullOrEmpty(stringValue))
                {
                    if (required) throw new ProviderException(string.Format("{0} is a required attribute for SolrSearchProvider.", key));                    
                }
                else if (!bool.TryParse(stringValue, out boolValue))
                {
                    throw new ProviderException(string.Format("{0} attribute could not be parsed.", key));
                }
                else result = boolValue;

                config.Remove(key);

                return result;
            }

            private int? TakeInt(NameValueCollection config, string key, bool required)
            {
                string stringValue = config[key];
                int intValue;
                int? result = null;

                if (string.IsNullOrEmpty(stringValue))
                {
                    if (required) throw new ProviderException(string.Format("{0} is a required attribute for SolrSearchProvider.", key));
                }
                else if (!int.TryParse(stringValue, out intValue))
                {
                    throw new ProviderException(string.Format("{0} attribute could not be parsed.", key));
                }
                else result = intValue;

                config.Remove(key);

                return result;
            }
            
            private void SetQueryBuilder(NameValueCollection config)
            {
                string queryBuilderTypeName = TakeString(config, "queryBuilderType", true);
                Type queryBuilderType;
                object queryBuilder;

                try
                {
                    queryBuilderType = Type.GetType(queryBuilderTypeName);
                }
                catch (Exception ex)
                {
                    throw new ProviderException("The query builder type could not be loaded.", ex);
                }

                try
                {
                    queryBuilder = Activator.CreateInstance(queryBuilderType);
                }
                catch (Exception ex)
                {
                    throw new ProviderException("The query builder could not be instantiated.", ex);
                }

                this.QueryBuilder = queryBuilder as ISearchQueryBuilder;
                if (queryBuilder == null)
                {
                    throw new ProviderException("The query builder type does not implement ISearchQueryBuilder.");
                }
            }
        }


        /// <summary>
        /// Initializes the provider.
        /// </summary>
        /// <param name="name">The friendly name of the provider.</param>
        /// <param name="config">A collection of the name/value pairs representing the provider-specific attributes specified in the configuration for this provider.</param>
        /// <exception cref="T:System.ArgumentNullException">
        /// The name of the provider is null.
        /// </exception>
        /// <exception cref="T:System.ArgumentException">
        /// The name of the provider has a length of zero.
        /// </exception>
        /// <exception cref="T:System.InvalidOperationException">
        /// An attempt is made to call <see cref="M:System.Configuration.Provider.ProviderBase.Initialize(System.String,System.Collections.Specialized.NameValueCollection)"/> on a provider after the provider has already been initialized.
        /// </exception>
        public override void Initialize(string name, NameValueCollection config)
        {
            Config = null;
            
            // validate the arguments and perform required inheritance actions.
            if (config == null) throw new ArgumentNullException("config");
            if (string.IsNullOrEmpty(name)) name = "SolrSearchProvider";
            base.Initialize(name, config);

            Config = new ProviderConfiguration(name, config);
        }

        #endregion


        #region Provider Public Methods

        /// <summary>
        /// Queries the Solr engine with the specified search criteria.
        /// </summary>
        /// <param name="applicationName">The application name.  May be null or empty if the provider is in shared core mode.</param>
        /// <param name="criteria">The search criteria.</param>
        /// <returns>An object representing the results of the search.</returns>
        public override ISearchResults Search(string applicationName, ISearchCriteria criteria)
        {            
            if (Config.DisableDateFilter)
            {
                CatalogEntrySearchCriteria catalogCriteria = criteria as CatalogEntrySearchCriteria;
                if (catalogCriteria != null) criteria = new SolrCatalogEntrySearchCriteria(catalogCriteria, Config.DisableDateFilter);
            }

            var query = (ISolrSearchQuery)Config.QueryBuilder.BuildQuery(criteria);

            if (Config.ShareCores)
            {
                string applicationQuery = new FieldQuery("_appname", applicationName).ToString();
                query.AddParameter("fq", applicationQuery);
            }

            var queryString = new StringBuilder();
            foreach (var param in query.Parameters)
            {
                if (queryString.Length > 0) queryString.Append('&');
                queryString
                    .Append(HttpUtility.UrlEncode(param.Key))
                    .Append('=')
                    .Append(HttpUtility.UrlEncode(param.Value));
            }

            if (query.Faceting != null)
            {
                var manager = query.Faceting as FacetManager;
                var hasFacetRequests = false;
                foreach (var facetField in query.Faceting.FacetFields)
                {
                    hasFacetRequests = true;
                    var facet = "&facet.field={!ex=tag" + facetField + "}";
                    queryString.Append(facet).Append(HttpUtility.UrlEncode(facetField));
                }

                foreach (var facetQuery in query.Faceting.FacetQueries)
                {
                    if (manager == null)
                    {
                        continue;
                    }
                    var field = manager.FacetQueryKeys.FirstOrDefault(x => x.Key.Equals(facetQuery)).Value;
                    hasFacetRequests = true;
                    var facet = "&facet.query={!ex=tag" + field.Key + " key=\"" + field.Value + "\"}";
                    queryString.Append(facet).Append(HttpUtility.UrlEncode(facetQuery));
                }

                if (hasFacetRequests)
                {
                    queryString.Append("&facet=true");
                }
            }

            SolrHttp.Response response;
            if (Config.QueryUsingPost)
            {
                var url = GetRequestUrl(applicationName, criteria.Scope, "select", null);
                response = SolrHttp.Post(url, queryString.ToString(), "application/x-www-form-urlencoded; charset=utf-8", Config.ResponseCache);

            }
            else
            {
                var url = GetRequestUrl(applicationName, criteria.Scope, "select", queryString.ToString());
                response = SolrHttp.Get(url, Config.ResponseCache);
            }

            var results = response.ProcessedObject as ISearchResults;
            if (results == null)
            {
                results = QueryResponseParser.Parse(criteria, query, response.Content, Config.FacetLocalizedFieldValuesOnly);
                response.ProcessedObject = results;
            }
            
            return results;
        }


        /// <summary>
        /// Adds or updates a document in the search index.  The operation may not be transmitted
        /// to the Solr server immediately.
        /// </summary>
        /// <param name="applicationName">The application name.</param>
        /// <param name="scope">The scope of the instance.</param>
        /// <param name="document">An object describing the document to be indexed.</param>
        public override void Index(string applicationName, string scope, ISearchDocument document)
        {
            UpdateBatch batch = GetBatch(applicationName, scope, true);
            batch.AddOrUpdate(document, Config.ShareCores ? applicationName : null);

            if (batch.BatchSize >= Config.MaximumBatchSize)
            {
                CommitBatch(GetRequestUrl(applicationName, scope, "update", null), batch);
            }
        }


        /// <summary>
        /// Removes documents matching the key/value pair from the search index.  The operation
        /// may not be transmitted to the Solr server immediately.
        /// </summary>
        /// <param name="applicationName">The application name.  May be null or empty if the provider is in shared core mode.</param>
        /// <param name="scope">The scope of the instance.</param>
        /// <param name="key">The name of the field to query for deletes on.</param>
        /// <param name="value">The value of the key field to delete.</param>
        /// <returns>Zero (meaningless).</returns>
        public override int Remove(string applicationName, string scope, string key, string value)
        {
            UpdateBatch batch = GetBatch(applicationName, scope, true);
            string query = new FieldQuery(key.ToLowerInvariant(), value).ToString();
            batch.DeleteByQuery(query);

            if (batch.BatchSize >= Config.MaximumBatchSize)
            {
                CommitBatch(GetRequestUrl(applicationName, scope, "update", null), batch);
            }

            return 0;
        }


        /// <summary>
        /// Removes all documents from the Solr instance, or all documents for the application
        /// in shared core mode.
        /// </summary>
        /// <param name="applicationName">The application name.  May be null or empty if the provider is in shared core mode.</param>
        /// <param name="scope">The scope of the instance.</param>
        public override void RemoveAll(string applicationName, string scope)
        {
            UpdateBatch batch = GetBatch(applicationName, scope, true);

            if (Config.ShareCores)
            {
                string query = new FieldQuery("_appname", applicationName).ToString();
                batch.DeleteByQuery(query);
            }
            else batch.DeleteByQuery("*:*", true);

            if (batch.BatchSize >= Config.MaximumBatchSize)
            {
                CommitBatch(GetRequestUrl(applicationName, scope, "update", null), batch);
            }
        }


        /// <summary>
        /// Sends and pending update and/or delete commands to the server.  This does not necessarily perform a commit.
        /// </summary>
        /// <remarks>
        /// <para>
        /// This send all pending operations to the Solr instance for a specified application name, even if they are in
        /// separate scopes.  In shared core mode, all pending updates will be sent.
        /// </para>
        /// <para>
        /// An actual commit may not be executed on the server.  If any additions/updates are included in the pending
        /// command list, a commitWithin directive will be sent.  If only deletes are included in the pending command
        /// list, a synchronous commit will be issued.
        /// </para>
        /// </remarks>
        /// <param name="applicationName">The application name to commit.</param>
        public override void Commit(string applicationName)
        {
            foreach (KeyValuePair<string, UpdateBatch> scopeAndBatch in GetBatches(applicationName))
            {
                string scope = scopeAndBatch.Key;
                string url = GetRequestUrl(applicationName, scope, "update", null);
                UpdateBatch batch = scopeAndBatch.Value;
                CommitBatch(url, batch);
            }
        }


        /// <summary>
        /// Commits the specified core and removes internal structures related to it.
        /// </summary>
        /// <param name="applicationName"></param>
        /// <param name="scope"></param>
        public override void Close(string applicationName, string scope)
        {
            UpdateBatch batch = GetBatch(applicationName, scope, false);
            if (batch != null)
            {
                string url = GetRequestUrl(applicationName, scope, "update", null);
                CommitBatch(url, batch);
                RemoveBatch(applicationName, scope);
            }
        }


        /// <summary>
        /// Generates an XML stub to paste into schema.xml describing Solr
        /// schema configuration for metafields.
        /// </summary>
        /// <param name="catalogContext">The metadata context to load catalog entry metaclass configuration from.</param>
        /// <returns>A string suitable for pasting into the supplied schema.xml.</returns>
        public string GenerateCatalogConfiguration(MetaDataContext catalogContext)
        {
            return SolrSchemaGenerator.GenerateCatalogConfiguration(catalogContext);
        }

        #endregion



        #region Private Implementation

        /// <summary>
        /// Formats a batch of commands, and posts the result to the Solr server.
        /// </summary>
        /// <param name="url">The update URL to post the command to.</param>
        /// <param name="batch">The batch to post.</param>
        private void CommitBatch(string url, UpdateBatch batch)
        {
            lock (batch.Lock)
            {
                try
                {
                    string document = batch.GetBatch(Config.CommitWithin);
                    if (document != null) SolrHttp.Post(url, document, "text/xml; charset=utf-8", null);
                }
                finally { batch.ClearBatch(); }
            }
        }


        /// <summary>
        /// Gets an object representing queued commands for the application and scope.
        /// If <paramref name="createIfNecessary"/> is true, a fresh batch object may be created.
        /// </summary>
        /// <remarks>
        /// This is overwrought and mildly inefficient, but the lack of a scope parameter in the API's
        /// commit call makes it difficult to store these in a more sane manner.
        /// </remarks>
        /// <param name="applicationName">The application name.  Ignored in shared core mode.</param>
        /// <param name="scope">The scope.</param>
        /// <param name="createIfNecessary">If true, a new batch may be created.</param>
        /// <returns>The specified batch.  If <paramref name="createIfNecessary"/> is false and the batch does not exist, null will be returned.</returns>
        private UpdateBatch GetBatch(string applicationName, string scope, bool createIfNecessary)
        {
            Dictionary<string, UpdateBatch> appBatches = null;
            UpdateBatch batch = null;

            if (Config.ShareCores) applicationName = string.Empty;

            if (!updateBatches.TryGetValue(applicationName, out appBatches) && createIfNecessary)
            {
                lock (updateBatchesLock)
                {
                    if (!updateBatches.TryGetValue(applicationName, out appBatches))
                    {
                        appBatches = new Dictionary<string, UpdateBatch>();
                        batch = new UpdateBatch();
                        appBatches.Add(scope, batch);
                        updateBatches.Add(applicationName, appBatches);
                    }
                }
            }

            if (appBatches != null && batch == null)
            {
                if (!appBatches.TryGetValue(scope, out batch) && createIfNecessary)
                {
                    lock (updateBatchesLock)
                    {
                        if (!appBatches.TryGetValue(scope, out batch))
                        {
                            batch = new UpdateBatch();
                            appBatches.Add(scope, batch);
                        }
                    }
                }
            }

            return batch;
        }


        /// <summary>
        /// Gets an anumeration of all the batches for an application, or all batches in shared core mode.
        /// </summary>
        /// <remarks>
        /// This is specifically to support the Commit call, which lacks a scope parameter.</remarks>
        /// <param name="applicationName">The application name.  Ignored in shared core mode.</param>
        /// <returns>
        /// A enumeration of the batches.
        /// The backing list for the enumeration is separate from the dictionary to avoid concurrent modification exceptions.
        /// </returns>
        private IEnumerable<KeyValuePair<string, UpdateBatch>> GetBatches(string applicationName)
        {
            if (Config.ShareCores) applicationName = string.Empty;

            List<KeyValuePair<string, UpdateBatch>> result;

            // prevent concurrent modification exceptions.
            lock (updateBatchesLock)
            {
                Dictionary<string, UpdateBatch> appBatches;
                if (updateBatches.TryGetValue(applicationName, out appBatches))
                {
                    result = appBatches.ToList();
                }
                else result = new List<KeyValuePair<string, UpdateBatch>>(0);
            }

            return result;
        }


        /// <summary>
        /// Removes a batch and all associated structures.
        /// </summary>
        /// <param name="applicationName">The application name for the batch.  Ignored in shared core mode.</param>
        /// <param name="scope">The scope.</param>
        private void RemoveBatch(string applicationName, string scope)
        {
            if (Config.ShareCores) applicationName = string.Empty;

            lock (updateBatchesLock)
            {
                Dictionary<string, UpdateBatch> appBatch;
                if (updateBatches.TryGetValue(applicationName, out appBatch))
                {
                    if (appBatch.Remove(scope))
                    {
                        if (updateBatches.Count == 0) updateBatches.Remove(applicationName);
                    }
                }
            }
        }


        /// <summary>
        /// Builds a url for a request.
        /// </summary>
        /// <param name="applicationName">The application name for the instance.  Ignored in shared core mode.</param>
        /// <param name="scope">The scope for the instance.</param>
        /// <param name="request">The Solr request url (supported operations for this provider are at "select" and "update").</param>
        /// <param name="query">The query parameters to send.</param>
        /// <returns>The absolute url describing the request.</returns>
        private string GetRequestUrl(string applicationName, string scope, string request, string query)
        {
            StringBuilder urlBuilder = new StringBuilder(Config.SolrServerUrl);

            if (!Config.ShareCores) urlBuilder.Append(applicationName).Append('.');
            urlBuilder.Append(scope);
            urlBuilder.Append('/').Append(request);

            if (!string.IsNullOrEmpty(query)) urlBuilder.Append('?').Append(query);

            return urlBuilder.ToString();
        }

        #endregion

        
    }
}
