﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Globalization;
using System.Linq;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Text.RegularExpressions;
using Mediachase.Commerce.Catalog;
using Mediachase.Search.Extensions;
using Mediachase.Search.Providers.Solr35.Select;

namespace Mediachase.Search.Providers.Solr35
{
    public class SolrSearchQueryBuilder : ISearchQueryBuilder
    {
        // matches character sequences -, +, !, ^, [, ], {, }, (, ), ", !, :, ;, \, ?, &&, and ||; and matches * unless it is the last non-whitespace character.
        private static readonly Regex SpecialCharacters = new Regex(@"([-[+!^{}()""~:;\\?]|\]|&&|\|\||\*[\s]*(?=[^\s]))", RegexOptions.Compiled);

        private static readonly string MinDateTimeString = DateTime.MinValue.ToString(SolrSearchProvider.DateTimeFormatString);
        private static readonly string MaxDateTimeString = DateTime.MaxValue.ToString(SolrSearchProvider.DateTimeFormatString);


        public static string EscapeValueString(string value)
        {
            string result = SpecialCharacters.Replace(value, "\\$1");
            return result.Contains(' ') ? string.Format("\"{0}\"", result) : result;
        }


        private class SolrSearchQuery : ISolrSearchQuery
        {
            private List<KeyValuePair<string, string>> parameters;

            public IEnumerable<KeyValuePair<string, string>> Parameters { get { return parameters; } }

            public IFacetManager Faceting { get; set; }

            public void AddParameter(string name, string value)
            {
                parameters.Add(new KeyValuePair<string, string>(name, value));
            }

            public SolrSearchQuery()
            {
                parameters = new List<KeyValuePair<string, string>>();
            }
        }


        /// <summary>
        /// Builds the query from the provided search criteria.
        /// </summary>
        /// <param name="criteria">The search criteria.</param>
        /// <returns>The query request object.</returns>
        public object BuildQuery(ISearchCriteria criteria)
        {
            var query = new SolrSearchQuery();

            query.AddParameter("q", BuildQueryString(criteria));
            
            var sortFields = criteria.Sort == null ? null : criteria.Sort.GetSort();
            if (sortFields != null && sortFields.Length > 0)
            {
                var sortBuilder = new StringBuilder();
                var first = true;
                foreach (var sort in sortFields)
                {
                    if (first) first = false;
                    else sortBuilder.Append(',');

                    sortBuilder.Append(sort.FieldName.ToLowerInvariant()).Append(sort.IsDescending ? " desc" : " asc");
                }

                query.AddParameter("sort", sortBuilder.ToString());
            }

            BuildActiveFieldQuery(criteria as AbstractSearchCriteria).ForEach(x => query.AddParameter(x.Key, x.Value));

            if (criteria.StartingRecord != 0) query.AddParameter("start", criteria.StartingRecord.ToString());
            query.AddParameter("rows", criteria.RecordsToRetrieve.ToString());

            query.Faceting = new FacetManager(criteria.Filters, criteria.Currency, criteria.MarketId);
            return query;
        }


        private string BuildQueryString(ISearchCriteria criteria)
        {
            var queryParts = new List<QueryPart>
            {
                new FieldQuery(BaseCatalogIndexBuilder.FieldConstants.Language, criteria.Locale.ToLower())
            };

            var solrCriteria = criteria as SolrCatalogEntrySearchCriteria;
            var catalogCriteria = solrCriteria != null ? solrCriteria.BaseCriteria : criteria as CatalogEntrySearchCriteria;
            if (catalogCriteria != null)
            {
                queryParts.Add(new ListQuery(BaseCatalogIndexBuilder.FieldConstants.Catalog, catalogCriteria.CatalogNames));
                queryParts.Add(new ListQuery(BaseCatalogIndexBuilder.FieldConstants.Node, catalogCriteria.CatalogNodes));
                queryParts.Add(new ListQuery(BaseCatalogIndexBuilder.FieldConstants.Outline, GetOutlines(catalogCriteria)));
                queryParts.Add(new ListQuery(BaseCatalogIndexBuilder.FieldConstants.MetaClass, catalogCriteria.SearchIndex));
                queryParts.Add(new ListQuery(BaseCatalogIndexBuilder.FieldConstants.ClassType, catalogCriteria.ClassTypes.Cast<string>().Select(s => s.ToLowerInvariant())));

                if (!criteria.IncludeInactive)
                {
                    var inactiveQuery = new ListQuery(BaseCatalogIndexBuilder.FieldConstants.IsActive, new StringCollection() { "true" });
                    queryParts.Add(inactiveQuery);
                }

                if (solrCriteria == null || !solrCriteria.DisableDateFilter)
                {
                    var queryStartAvailbleDateText = new StringBuilder();
                    queryStartAvailbleDateText.Append("(");

                    queryStartAvailbleDateText.Append("(");
                    queryStartAvailbleDateText.Append(new FieldQuery(BaseCatalogIndexBuilder.FieldConstants.AllowPreorder, catalogCriteria.IncludePreorderEntry.ToString()).ToString());
                    queryStartAvailbleDateText.Append(" AND ");
                    queryStartAvailbleDateText.Append(new RangeQuery(BaseCatalogIndexBuilder.FieldConstants.PreorderAvailableDate, MinDateTimeString, catalogCriteria.StartDate.ToString(SolrSearchProvider.DateTimeFormatString)).ToString());
                    queryStartAvailbleDateText.Append(")");

                    queryStartAvailbleDateText.Append(" OR ");

                    // TODO: Change the date filter in ProductSearchProvider.cs when changing this. 
                    queryStartAvailbleDateText.Append(new RangeQuery(BaseCatalogIndexBuilder.FieldConstants.StartDate, MinDateTimeString, catalogCriteria.StartDate.ToString(SolrSearchProvider.DateTimeFormatString)).ToString());
                    queryStartAvailbleDateText.Append(")");

                    var queryStartAvailbleDate = new VerbatimQuery(queryStartAvailbleDateText.ToString());
                    queryParts.Add(queryStartAvailbleDate);

                    queryParts.Add(new RangeQuery(BaseCatalogIndexBuilder.FieldConstants.EndDate, catalogCriteria.EndDate.ToString(SolrSearchProvider.DateTimeFormatString), MaxDateTimeString));
                }

                // Add market filter _ExcludedCatalogEntryMarkets
                if (!string.IsNullOrEmpty(criteria.MarketId.Value))
                    queryParts.Add(new FieldQuery(string.Format("-{0}", ExcludedCatalogEntryMarketsField.FieldName.ToLower()), criteria.MarketId.Value.ToLowerInvariant()));

                if (!string.IsNullOrEmpty(catalogCriteria.SearchPhrase))
                {
                    if (catalogCriteria.IsFuzzySearch)
                    {
                        // TODO: Verify this is correct functionality (issue 1939).
                        // TODO: When verified, move this out to an AddXxxQuery(state, ...) method.
                        var searchPhraseBuilder = new StringBuilder();
                        foreach (string keyword in catalogCriteria.SearchPhrase.Split(' ').Where(s => !string.IsNullOrEmpty(s)))
                        {
                            searchPhraseBuilder.AppendFormat("{0}~{1}", keyword, catalogCriteria.FuzzyMinSimilarity.ToString(CultureInfo.InvariantCulture));
                        }

                        if (searchPhraseBuilder.Length > 0) queryParts.Add(new VerbatimQuery(searchPhraseBuilder.ToString()));
                    }
                    else
                    {
                        queryParts.Add(new VerbatimQuery(catalogCriteria.SearchPhrase));
                    }
                }
            }

            // NOTE: Add any additional parts BEFORE the catalogCriteria.SearchPhrase handling.

            var prefix = new StringBuilder();
            var body = new StringBuilder();
            var first = true;

            foreach (var part in queryParts.Where(p => p.HasContent()))
            {
                if (!first)
                {
                    prefix.Append('(');
                    body.Append(" AND ");
                }

                part.AppendTo(body);

                if (!first) body.Append(')');
                else first = false;
            }

            return prefix.Append(body).ToString();
        }

        private static List<KeyValuePair<String,String>> BuildActiveFieldQuery(AbstractSearchCriteria criteria)
        {
            if (criteria == null)
            {
                return null;
            }
            var fqList = new List<KeyValuePair<String,String>>();
            var query = new StringBuilder();
            var activeFilterFields = criteria.ActiveFilterFields;
            foreach (var field in activeFilterFields)
            {
                var fieldName = field.ToLowerInvariant();
                var tag = "{!tag=tag" + fieldName + "}";
                var values = criteria.GetSelectedSearchFilterValues(field).ToList();
                if (values.Count == 1)
                {
                    var value = values.First();
                    SimpleValue simpleValue;
                    RangeValue rangeValue;
                    PriceRangeValue priceRangeValue;

                    if ((simpleValue = (value as SimpleValue)) != null)
                    {
                        fqList.Add(new KeyValuePair<string, string>("fq", tag + new FieldQuery(fieldName, simpleValue.value)));
                    }
                    else if ((rangeValue = (value as RangeValue)) != null)
                    {
                        fqList.Add(new KeyValuePair<string, string>("fq", tag + new RangeQuery(fieldName,
                            rangeValue.lowerbound, rangeValue.upperbound,
                            rangeValue.lowerboundincluded, rangeValue.upperboundincluded)));
                    }
                    else if ((priceRangeValue = (value as PriceRangeValue)) != null)
                    {
                        if (priceRangeValue.currency.Equals(criteria.Currency, StringComparison.OrdinalIgnoreCase))
                        {
                            fqList.Add(new KeyValuePair<string, string>("fq", tag +new RangeQuery(fieldName + priceRangeValue.currency,
                                priceRangeValue.lowerbound, priceRangeValue.upperbound,
                                priceRangeValue.lowerboundincluded, priceRangeValue.upperboundincluded)));
                        }
                    }
                }
                else
                {
                    var value = values.First();
                    if (value as SimpleValue != null)
                    {
                        fqList.Add(new KeyValuePair<string, string>("fq", tag + new MultiValueFieldQuery(fieldName, values.OfType<SimpleValue>().Select(x => x.value))));
                    }
                    else if (value as RangeValue != null)
                    {
                        var ranges = values.OfType<RangeValue>().
                            Select(val => new RangeQuery(fieldName, val.lowerbound, val.upperbound, val.lowerboundincluded, val.upperboundincluded));
                        fqList.Add(new KeyValuePair<string, string>("fq", tag + String.Join(" OR ", ranges)));
                    }
                    else if (value as PriceRangeValue != null)
                    {
                        var ranges = values.OfType<PriceRangeValue>().
                            Where(x => x.currency.Equals(criteria.Currency, StringComparison.OrdinalIgnoreCase)).
                            Select(val => new RangeQuery(fieldName + val.currency, val.lowerbound, val.upperbound, val.lowerboundincluded, val.upperboundincluded));
                        fqList.Add(new KeyValuePair<string, string>("fq", tag + String.Join(" OR ", ranges)));
                    }
                }
               
            }
            return fqList;
        }

        /// <summary>
        /// Get outlines from catalogentrysearch criteria by removing asterisk .
        /// </summary>
        /// <param name="criteria">The catalog entry search criteria.</param>
        /// <returns></returns>
        private StringCollection GetOutlines(CatalogEntrySearchCriteria criteria)
        {
            var results = new StringCollection();
            foreach (var item in criteria.Outlines)
            {
                results.Add(item.EndsWith("*")
                                ? item.Substring(0, item.Length - 1)
                                : item);
            }
            return results;
        }
    }
}
