﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Xml.Linq;

namespace Mediachase.Search.Providers.Solr35.Update
{
    internal static class DocumentSerializer
    {
        #region Safe Handlers for dodgy ISearchField.Values nested arrays

        private static readonly object[] NoValues = new object[0];

        private static IEnumerable GetValues(ISearchField field)
        {
            if (field.Values != null) return GetValues(field.Values);
            else if (field.Value != null) return new object[] { field.Value }.AsEnumerable();
            else return NoValues.AsEnumerable();
        }

        private static IEnumerable GetValues(Array array)
        {
            IEnumerable result = null;

            if (array.Length == 1)
            {
                Array nestedArray = array.GetValue(0) as Array;
                if (nestedArray != null) result = GetValues(nestedArray);
            }
            
            if (result == null) result = EnumerateArray(array);

            return result;
        }

        private static IEnumerable EnumerateArray(Array array)
        {
            for (int i = 0; i < array.Length; ++i)
            {
                object value = array.GetValue(i);
                if (value != null) yield return value;
            }
        }

        #endregion



        public static XElement Serialize(ISearchDocument document, string appName)
        {
            XElement docElement = new XElement("doc");

            if (!string.IsNullOrEmpty(appName))
            {
                docElement.Add(new XElement("field", new XAttribute("name", "_appname"), new XText(appName)));
            }

            for (int i = 0; i < document.FieldCount; ++i)
            {
                ISearchField field = document[i];
                XAttribute name = new XAttribute("name", field.Name.ToLowerInvariant());
                
                foreach (object value in GetValues(field))
                {
                    XText valueNode = SerializeValue(value);
                    docElement.Add(new XElement("field", name, valueNode));
                }
            }            

            return docElement;
        }


        private static XText SerializeValue(object value)
        {
            string text = value as string;
            if (text == null)
            {
                Type type = value.GetType();
                Type valueType = Nullable.GetUnderlyingType(type) ?? type;
                
                if (valueType == typeof(DateTime))
                {
                    text = ((DateTime)value).ToString(SolrSearchProvider.DateTimeFormatString);
                }
                else if (valueType == typeof(bool))
                {
                    text = ((bool)value) ? "true" : "false";
                }
                else 
                {
                    try 
                    {
                        // TODO: should this be memoized? probably, but it may be tricky to accomplish that
                        // in the current provider model without resorting to a full static cache object
                        text = TypeDescriptor.GetConverter(type).ConvertToInvariantString(value);
                    }
                    catch (NotSupportedException)
                    {
                        throw new Exception("Cannot serialize field value of type " + value.GetType());
                    }
                }
            }

            return new XText(text);
        }
    }
}
