﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;

namespace Mediachase.Search.Providers.Solr35.Update
{
    internal class UpdateBatch
    {
        private object batchLock;
        private List<string> pendingDeletes;
        private List<XElement> pendingAdds;

        public int BatchSize
        {
            get { return pendingDeletes.Count + pendingAdds.Count; }
        }

        public object Lock
        {
            get { return batchLock; }
        }

        public UpdateBatch()
        {
            batchLock = new object();
            pendingDeletes = new List<string>();
            pendingAdds = new List<XElement>();
        }

        
        public void DeleteByQuery(string query)
        {
            this.DeleteByQuery(query, false);
        }


        public void DeleteByQuery(string query, bool clearOtherDeletes)
        {
            if (clearOtherDeletes) pendingDeletes.Clear();
            pendingDeletes.Add(query);
        }


        public void AddOrUpdate(ISearchDocument document, string appName)
        {
            pendingAdds.Add(DocumentSerializer.Serialize(document, appName));
        }


        /// <summary>
        /// Serializes all indexing commands in the batch into an xml document suitable for posting.
        /// </summary>
        /// <param name="commitWithin">
        /// If not null, and there are any add commands, a commitWithin directive is sent.
        /// If not null, and there are only deletes, a commit command is added to the document.</param>
        /// If null, all commit handling is left to the called.
        /// <returns>An xml string suitable for posting, or null if there are no commands.</returns>
        public string GetBatch(int? commitWithin)
        {
            bool hasDeletes = pendingDeletes.Count > 0;
            bool hasAdds = pendingAdds.Count > 0;

            if (!hasDeletes && !hasAdds) return null;


            XElement updateElement = new XElement("update");
            
            if (hasDeletes) updateElement.Add(GetDeleteElement());
            
            if (hasAdds) updateElement.Add(GetAddElement(commitWithin));
            else if (commitWithin.HasValue)
            {
                // if there are only deletes, and a commitWithin value is specified, the document is expected to have
                // some commit directive.  since commitWithin is not supported without adds, add an explicit commit.
                updateElement.Add(new XElement("commit"));
            }

            return updateElement.ToString();
        }


        public void ClearBatch()
        {
            pendingDeletes.Clear();
            pendingAdds.Clear();
        }


        private XElement GetDeleteElement()
        {
            XElement deleteElement = new XElement("delete");
            
            foreach (string query in pendingDeletes)
            {
                deleteElement.Add(new XElement("query", query));
            }

            return deleteElement;
        }


        private XElement GetAddElement(int? commitWithin)
        {
            XElement addElement = new XElement("add");
            if (commitWithin.HasValue) addElement.Add(new XAttribute("commitWithin", commitWithin.Value.ToString()));
            foreach (XElement doc in pendingAdds) addElement.Add(doc);
            return addElement;
        }
    }
}
